fs            = require "fs"
os            = require "os"
nodePath      = require "path"
https         = require "https"
shell         = require "shelljs"
colors        = require "colors"
coffee        = require "coffee-script"
{exec, spawn} = require "child_process"
{ask, logger, compileDebug} = require "../lib/utils"
log           = logger "kdapp"

# Make sure to work with node's old versions
unless fs.existsSync
  fs.existsSync = nodePath.existsSync

# Utilities
manifest = (username, name)->
  JSON.stringify
    devMode       : true
    version       : "0.1"
    name          : name
    identifier    : "com.koding.apps.#{name}"
    path          : "~/Applications/#{name}.kdapp"
    homepage      : "#{username}.koding.com/#{name}"
    author        : "#{username}"
    repository    : "git://github.com/#{username}/#{name}.kdapp.git"
    description   : "#{name} : a Koding application created with the KD cli."
    category      : "web-app"
    source        :
      blocks      :
        app       :
          files   : [
            "./index.coffee"
          ]
      stylesheets : [
        "./resources/style.css"
      ]
    options       :
      type        : "tab"
    icns          :
      128         : "./resources/icon.128.png"
  , null, 2

escapeshell = (cmd)=> 
  cmd = cmd.replace /(["\s'$`\\])/g, '\\$1'
  "\"#{cmd}\""

# Module
module.exports = class App
  constructor: ({@config})->

  pistachios: /\{(\w*)?(\#\w*)?((?:\.\w*)*)(\[(?:\b\w*\b)(?:\=[\"|\']?.*[\"|\']?)\])*\{([^{}]*)\}\s*\}/g

  compile: (path)->

    {argv: {verbose}} = @options
      .usage("Compiles applications")
      .alias("v", "verbose")
      .describe("v", "Show error details")

    appPath = path or process.cwd()

    for manifestName in [".manifest", "manifest.json"]
      manifestPath = (nodePath.join appPath, manifestName)
      break  if fs.existsSync manifestPath

    try
      manifest = JSON.parse fs.readFileSync manifestPath
    catch err
      if err.errno is 34
        log "Manifest file does not exist '#{manifestPath}'", "red"
        process.exit 34
      else
        log "Manifest file seems corrupted '#{manifestPath}'", "red"
        log err, "red"
        process.exit 3

    files = manifest.source.blocks.app.files
    source = ""

    for file in files
      file = nodePath.normalize (nodePath.join path, file)  if path
      log "compiling #{file}...", "magenta"  if verbose

      if /\.coffee/.test file
        data = fs.readFileSync file
        try
          compiled = coffee.compile data.toString(), bare: true
        catch error
          log error, "red"
          if verbose 
            log compileDebug(file, data, error), "red"
          process.exit 1

      else if /\.js/.test file
        try
          compiled = fs.readFileSync(file).toString()
        catch error
          log error, "red"
          process.exit 1

      block = """
      /* BLOCK STARTS: #{file} */
      #{compiled}
      """
      block = block.replace @pistachios, (pistachio)-> pistachio.replace /\@/g, 'this.'
      source += block

    log "creating index.js file...", "magenta"  if verbose
    mainSource = """
    /* Compiled by KD on #{(new Date()).toString()} */
    (function() {
    /* KDAPP STARTS */
    #{source}
    /* KDAPP ENDS */
    }).call();
    """
    fs.writeFileSync (nodePath.join appPath, "index.js"), mainSource
    log "Application has been compiled!", "green"

  create: (name, path)->

    unless name 
      return log "You have to define a name to create an app!", "red"

    {argv: {sync}} = @options
      .usage("Creates a Koding Application template")
      .alias("s", "sync")
      .describe("s", "Sync kite after creation")

    appPath = nodePath.resolve path or process.cwd()

    unless @config.username
      return log """
      I don't know who you are.

      Please run: kd config username
      """, "red"

    if name.match /[^\w]/
      return log "You mustn't use special chars in kite name."

    appDir = nodePath.join appPath, "#{name}.kdapp"

    # That's used to be shame, now fixed. :/
    # Bash file to run.
    log "Creating #{name}.kdapp..."
    shell.mkdir "-p", appDir
    shell.mkdir "-p",(nodePath.join appDir, "resources")
    fs.writeFileSync (nodePath.join appDir, "manifest.json"), manifest(@config['username'], name)
    fs.writeFileSync (nodePath.join appDir, "ChangeLog"), ""
    fs.writeFileSync (nodePath.join appDir, "README"), "# #{name}.kdapp#{os.EOL}"
    fs.writeFileSync (nodePath.join appDir, "index.coffee"), "do ->#{os.EOL}  # appView.addSubView"
    fs.writeFileSync (nodePath.join appDir, "resources", "style.css"), ""
    icon = fs.createWriteStream (nodePath.join appDir, "resources", "icon.128.png")
    request = https.get "https://koding.com/images/default.app.thumb.png", (response)-> response.pipe icon
    request.on "error", ->
      log "Some resources couldn't be downloaded. Check your internet connection.", "red"
      process.exit 1
    log """
    You have created a new app.

    cd #{nodePath.relative process.cwd(), appDir}
    kd app compile
    """, "green"
    if sync
      process.chdir appDir
      @sync()


  sync: (path)->

    appPath = path or process.cwd()

    unless appPath.match /\.kdapp$/
      return log """
      You are not in an application directory. Application directory names
      must end with `.kdapp` extension.

      Like that example:

      kd app create appname
      cd appname.kdapp
      """, "red"


    try
      manifest = JSON.parse fs.readFileSync nodePath.join appPath, "manifest.json"
    catch err
      if err.errno is 34
        log "Manifest file does not exist '#{manifestPath}'", "red"
        process.exit 34
      else
        log "Manifest file seems corrupted '#{manifestPath}'", "red"
        log err, "red"
        process.exit 3

    ask "Please enter your Koding password:",
      hidden: true
      callback: (password)=>
        log "Connecting your Koding filesystem, please wait..."

        ftps = require "ftps"
        connection = new ftps
          host      : "ftps.koding.com"
          username  : @config['username']
          password  : password
          protocol  : "ftps"
        connection
        .raw("set ssl:verify-certificate no")
        .cd("Applications")
        .raw("mirror -Ren #{appPath}")
        .exec (err, {_err, data})->
          if err or _err
            log err
            process.exit 1
          else
            log data
            process.exit 0
