fs                  = require "fs"
os                  = require "os"
nodePath            = require "path"
url                 = require "url"
coffee              = require "coffee-script"
colors              = require "colors"
request             = require "request"
http                = require "http"
https               = require "https"
Progress            = require "progress"
shell               = require "shelljs"
{spawn, exec}       = require "child_process"
{ask, logger, compileDebug}   = require "../lib/utils"
{template, unique}  = require "underscore"
log                 = logger "kite"

# Make sure to work with node's old versions
unless fs.existsSync
  fs.existsSync = nodePath.existsSync

# Utilities
manifest = (username, name)->
  JSON.stringify
    devMode       : true
    version       : "0.1"
    name          : name
    identifier    : "com.koding.kites.#{name}"
    path          : "~/Kites/#{name}.kite"
    homepage      : "#{username}.koding.com/kites/#{name}"
    author        : "#{username}"
    repository    : "git://github.com/#{username}/#{name}.kite.git"
    description   : "#{name} : a Koding kite created with the KD cli."
    category      : "web-api"
    source        :
      blocks      :
        kite      :
          files   : [
            "./index.coffee"
          ]
      scripts : {
        "run": "node index.js"
      }
    options       :
      privacy     : "private"
    icns          :
      128         : "./resources/icon.128.png"
    dependencies  : [
      # "./kites/#{name}/helloworld"
    ]
  , null, 2

# Module
module.exports = class Kite

  alias:
    "new"     : "create"
    "i"       : "install"
    "depends" : "install"

  help:"""
  Kites are simply web services for Koding.
  You can share your kites over internet.
  """

  constructor: ({@config})->
    @config.KODING_ROOT or= "https://koding.com"

  compile: (path)->

    {argv: {verbose}} = @options
      .usage("Compiles kites")
      .alias("v", "verbose")
      .describe("v", "Show error details")

    kitePath = path or process.cwd()

    for manifestName in [".manifest", "manifest.json"]
      manifestPath = (nodePath.join kitePath, manifestName)
      break  if fs.existsSync manifestPath

    try
      manifest = JSON.parse fs.readFileSync manifestPath
    catch err
      if err.errno is 34
        log "Manifest file does not exist '#{manifestPath}'", "red"
        process.exit 34
      else
        log "Manifest file seems corrupted '#{manifestPath}'", "red"
        log err, "red"
        process.exit 3

    files = manifest.source.blocks.kite.files
    source = ""

    for file in files
      file = nodePath.normalize (nodePath.join path, file)  if path
      log "compiling #{file}...", "magenta"  if verbose

      if /\.coffee/.test file
        data = fs.readFileSync file
        try
          compiled = coffee.compile data.toString(), bare: true
        catch error
          log error, "red"
          if verbose 
            log compileDebug(file, data, error), "red"
            process.exit 1

      else if /\.js/.test file
        try
          compiled = fs.readFileSync(file).toString()
        catch error
          console.log error

      block = """
      /* BLOCK STARTS: #{file} */
      #{compiled}
      """
      block = block.replace @pistachios, (pistachio)-> pistachio.replace /\@/g, 'this.'
      source += block

    log "creating index.js file...", "magenta"  if verbose
    mainSource = """
    /* Compiled by KD on #{(new Date()).toString()} */
    (function() {
    /* KITE STARTS */
    #{source}
    /* KITE ENDS */
    }).call();
    """
    fs.writeFileSync (nodePath.join kitePath, "index.js"), mainSource
    log "Kite has been compiled!", "green"

  create: (name, path)->
    
    kitePath = path or process.cwd()

    unless name
      log "You have to define a name for your kite.", "red"
      process.exit 1

    if name.match /[^\w]/
      log "You mustn't use special chars in kite name.", "red"
      process.exit 1

    kiteDir = nodePath.join kitePath, "#{name}.kite"

    # Templates
    indexFile = template (fs.readFileSync nodePath.join __dirname, "templates", "kite", "index.tpl.coffee").toString()
    keyFile = template (fs.readFileSync nodePath.join __dirname, "templates", "kite", "key.tpl.coffee").toString()
    
    log "Creating #{name}.kite..."
    shell.mkdir "-p", kiteDir
    shell.mkdir "-p",(nodePath.join kiteDir, "resources")
    shell.mkdir "-p",(nodePath.join kiteDir, "lib")
    shell.mkdir "-p",(nodePath.join kiteDir, "test")
    shell.mkdir "-p",(nodePath.join kiteDir, "node_modules")
    fs.writeFileSync (nodePath.join kiteDir, "manifest.json"), manifest(@config['username'], name)
    fs.writeFileSync (nodePath.join kiteDir, "ChangeLog"), ""
    fs.writeFileSync (nodePath.join kiteDir, "README"), "# #{name}.kite#{os.EOL}"
    fs.writeFileSync (nodePath.join kiteDir, "lib", "key.coffee"), keyFile()
    fs.writeFileSync (nodePath.join kiteDir, "index.coffee"), indexFile {name: name, username: @config.username, mail: @config.mail}
    fs.writeFileSync (nodePath.join kiteDir, "test", "test.coffee"), ""
    icon = fs.createWriteStream (nodePath.join kiteDir, "resources", "icon.128.png")
    request = https.get "https://koding.com/images/default.app.thumb.png", (response)-> response.pipe icon
    request.on "error", (error)->
      log "Some resources couldn't be downloaded. Check your internet connection.", "red"
      process.exit 1
    npmInstall = """
    cd #{kiteDir}
    npm i kd-kite kd-rope mocha
    """

    tmpFile = nodePath.join os.tmpDir(), "koding.kd.kite.create.#{Math.random()}"
    fs.writeFileSync tmpFile, npmInstall

    install = spawn process.env.SHELL, [tmpFile]

    progress = new Progress "#{colors.yellow 'kite:'} Installing dependencies: [:bar] :percent :etas", 
      total: 200
      incomplete: " "
      width: 20

    install.stdout.on "data", (data)-> do progress.tick
    install.stderr.on "data", (data)-> do progress.tick

    install.on "close", ->
      progress.tick progress.total - progress.curr # complete the blank.
      process.stdout.write os.EOL
      log """
      Your kite created successfully.
      Now enter into kite directory and run the kite with the following commands:

      cd #{nodePath.relative process.cwd(), kiteDir}
      kd kite run
      """, "green"
      process.exit()

  install: (repo, path)->

    kitePath = path or process.cwd()

    unless shell.which "git"
      log "You need git to install dependencies.", "red"
      process.exit 1

    for manifestName in [".manifest", "manifest.json"]
      manifestPath = (nodePath.join kitePath, manifestName)
      break  if fs.existsSync manifestPath

    try
      manifest = JSON.parse fs.readFileSync manifestPath
    catch err
      if err.errno is 34
        log "Manifest file does not exist '#{manifestPath}'", "red"
        process.exit 34
      else
        log "Manifest file seems corrupted '#{manifestPath}'", "red"
        log err, "red"
        process.exit 3

    {dependencies} = manifest
    dependencies = [repo]  if repo
    dependencies = unique(dependencies).filter (dependency)-> dependency

    for dependency, i in dependencies
      log dependency
      repoUrl = url.parse dependency
      unless repoUrl.hostname
        repoUrlString = url.format
          protocol: 'https'
          host: 'github.com'
          pathname: repoUrl.pathname
        dependencies[i] = repoUrlString

    for dependency in dependencies
      {pathname} = url.parse dependency
      clonePath = (nodePath.join kitePath, "kites", pathname)
      unless /\.kite$/.test clonePath
        log """
        The repo name doesn't have '.kite' extension, maybe it's not a kite.
        If it's not a real kite, 'kd kite run' won't work.
        """, "magenta"

      log "Installing: #{dependency}...", "blue"
      shell.mkdir "-p", clonePath
      shell.exec "git clone --recursive -q #{dependency} #{clonePath}", (err, res)->
        log "#{pathname} installed.", "green"
        # TODO: install sub dependencies!
      manifest.dependencies.push (nodePath.relative kitePath, clonePath)
      manifest.dependencies = unique manifest.dependencies
      fs.writeFileSync manifestPath, (JSON.stringify manifest, null, 2)

  run: (path)->

    {argv: {verbose}} = @options
      .usage("Runs kites")
      .alias("v", "verbose")
      .describe("v", "Show error details")

    kitePath = path or process.cwd()

    @compile kitePath

    for manifestName in [".manifest", "manifest.json"]
      manifestPath = (nodePath.join kitePath, manifestName)
      break  if fs.existsSync manifestPath

    try
      manifest = JSON.parse fs.readFileSync manifestPath
    catch err
      if err.errno is 34
        log "Manifest file does not exist '#{manifestPath}'", "red"
        process.exit 34
      else
        log "Manifest file seems corrupted '#{manifestPath}'", "red"
        log err, "red"
        process.exit 3

    log "Attempting to run #{manifest.name}", "inverse"

    unless manifest.apiAddress
      log "API address not found: Defining api address to #{@config.KODING_ROOT} on manifest.", "magenta"
      @manifest "apiAddress", @config.KODING_ROOT

    if manifest.dependencies?.length > 0
      log "Running dependencies...", "magenta"
      for dependencyPath in manifest.dependencies
        log "Running dependency #{nodePath.basename dependencyPath}"
        process.chdir nodePath.join kitePath, dependencyPath
        @run nodePath.join kitePath, dependencyPath
        

    process.chdir kitePath
    kiteSource = nodePath.join kitePath, "index.js"
    log "Running #{kiteSource}"  if verbose

    if fs.existsSync kiteSource
        error = ""
        kiteRuntime = spawn shell.which("node"), [kiteSource]
        kiteRuntime.stdout.on "data", (data)->
          log data.toString().trim(), "blue"
        kiteRuntime.stderr.on "data", (data)->
          error += data.toString()
        kiteRuntime.on "exit", (code)->
          if code > 0
            log "Kite couldn't run. To see details user `--verbose`", "red"
            if verbose then log error, "red"
          process.exit code
    else
      log "The index.js doesn't exist. You should run `kd kite compile`", "red"
      process.exit 1

  server: (path)->

    kitePath = path or process.cwd()

    @run kitePath
    kite = require nodePath.join kitePath, "index.js"
    httpd = http.createServer (req, res)->
      {kite: {api}} = kite
      res.writeHead 200,
        'Content-Type': 'application/json'
        'Access-Control-Allow-Origin': '*'

      api["helloWorld"] name: "fka", (error, result)->
        res.end JSON.stringify {error, result}
    log "Server is running on port 9898", "green"
    httpd.listen 9898

  manifest: (key, value, path)->

    kitePath = path or process.cwd()

    for manifestName in [".manifest", "manifest.json"]
      manifestPath = (nodePath.join kitePath, manifestName) 
      break  if fs.existsSync manifestPath

    try
      manifest = JSON.parse fs.readFileSync manifestPath
    catch err
      if err.errno is 34
        log "Manifest file does not exist '#{manifestPath}'", "red"
        process.exit 34
      else
        log "Manifest file seems corrupted '#{manifestPath}'", "red"
        log err, "red"
        process.exit 3

    unless value
      console.log eval "manifest.#{key}"
    else
      eval "manifest.#{key} = value"
      manifestData = JSON.stringify manifest, null, 2
      fs.writeFileSync manifestPath, manifestData

  test: (path)->

    kitePath = path or process.cwd()

    kiteTestFile = nodePath.join kitePath, "test", "test.coffee"
    mocha = nodePath.join kitePath, "node_modules", "mocha", "bin", "mocha"
    unless fs.existsSync mocha
      log "Mocha is not installed. To install mocha, run `npm install mocha`", "red"
      process.exit 1
    if fs.existsSync kiteTestFile
      exec "coffee -c #{kiteTestFile}", ->
        child = spawn mocha, [kiteTestFile]
        child.stdout.on "data", (data)->
          process.stdout.write data.toString()
    else
      log "The test file doesn't exist.", "red"
      process.exit 1
